MODULE GENINPUT
USE UTILS
use genout_tools
use global_vars
use genout_tools2
!
IMPLICIT NONE
!

CONTAINS
!
!*************************
SUBROUTINE GENINPUT_READNODES(filename)
!*************************
! Input
CHARACTER*(*) filename
! locals
integer :: unit,i
! Implementation
CALL getfreeunit(unit)
OPEN(unit,FILE=filename)
do i=1,5
  read(unit,*)
enddo
DO WHILE (.NOT.EOF(unit))
  call inc_nodes_array()
  READ(unit,*) nodes(size(nodes))
ENDDO
CLOSE(unit)
!
END SUBROUTINE
!
!*************************
SUBROUTINE GENINPUT_READMEMBERS(filename)
!*************************
! Input
CHARACTER*(*) filename
! locals
TYPE (TMEMBER), POINTER :: m
integer					:: nodeid1,nodeid2
logical					:: found
integer					:: i
integer :: unit
! Implementation

CALL getfreeunit(unit)
OPEN(unit,FILE=filename)
do i=1,5
  read(unit,*)
enddo
DO WHILE (.NOT.EOF(unit))
  call inc_members_array()
  m => MEMBERS(size(members))
  READ(unit,*) m%ID,             &
             nodeid1,nodeid2, &
!             m%NODE1, m%NODE2, &
			 m%D1, m%D2,       &
			 m%T1, m%T2 ,       &
			 m%marine_growth, &
			 m%Dif
  !find node related to member nodeid1
  found=.false.
  do i=1,size(nodes)
    if (nodes(i)%id.eq.nodeid1) then
	  m%node1=i
	  found=.true.
	  exit
	endif
  enddo
  if (found.eq..false.) write (0,*) '*** ERROR *** Node not found'
  !find node related to member nodeid2
  found=.false.
  do i=1,size(nodes)
    if (nodes(i)%id.eq.nodeid2) then
	  m%node2=i
	  found=.true.
	  exit
	endif
  enddo
  if (found.eq..false.) write (0,*) '*** ERROR *** Node not found'
ENDDO
CLOSE(unit)
!
END SUBROUTINE
!
!*************************
SUBROUTINE GENINPUT_READCONNECT(filename)
use dfport
use genout_tools2
!*************************
! Input
CHARACTER*(*) filename
! locals
!
Type (Tmasterfile),pointer :: masterfile

Type (Tline) :: line
logical,save :: eof                                ! End Of File
               
Type (Tword) :: words
logical                    :: init,&
                              err,&
							  nbodies_control

real*8, dimension(maxwords) :: rea

integer :: error, info_array(12)

err=.false.
nbodies_control=.false.

!call getarg(1,masterfile%name)                     ! filename from HAWC2 command parameter retrieved
allocate(masterfile)
masterfile%name = filename
call getfreeunit(masterfile%unit)           

error=stat(masterfile%name,info_array)
if (error/=0) then
  GOTO 10
else 
  open (unit=masterfile%unit, file=masterfile%name,err=10)  ! commandfile is opened
endif

init=.true.
eof=.false.
do while (eof==.false.)
    call readline(masterfile,line,init)          ! line is read
    call getwords(line%line,words,masterfile)                 ! words in line is retrieved

!	if (trim(words%word_vec(1))=='nbodies') then
!      call get_parameters(words,2,1,rea,err)
!      if (err==.true.) exit
!      nbodies=nint(rea(1))
!      nbodies_control=.true.
!      allocate(bodies(nbodies))

!    elseif (trim(words%word_vec(1))=='begin') then                   
    if (trim(words%word_vec(1))=='begin') then                   
      select case (trim(words%word_vec(2)))
        case ('properties')                              ! div. input
          call properties_input(masterfile)

        case ('body')                              ! body input
          call body_input(masterfile)

        case ('fix')                              ! constraint input
          call fix_input(masterfile)

        case ('')
          continue

        case default
          write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
      end select

    else if (trim(words%word_vec(1))=='') then
      continue
    else if (trim(words%word_vec(1))=='exit') then       ! End of file
      eof=.true.
      exit
    else
      write(0,*) '*** ERROR *** Command unknown and ignored. Line number: ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    endif
enddo

close (unit=masterfile%unit)

return
10 write(*,*) 'Error opening file: ',trim(masterfile%name)
stop

END SUBROUTINE
!
Subroutine properties_input(masterfile)
use genout_tools

Type (Tmasterfile), pointer :: masterfile

Type (Tline) :: line
Type (Tword) :: words

integer :: int
real*8, dimension(maxwords) :: rea

logical :: init                     , &
           out                      , &
           name_control            , &
           nmembers_control            , &
           memberid_control            , &
           all_ok, &
           err

init                          =.false.
out                           =.false.
name_control                  =.false.
nmembers_control                 =.false.
memberid_control               =.false.
err=.false.
all_ok=.false.

  do while (out==.false.)
    call readline(masterfile,line,init)        ! line is read
    call getwords(line%line,words,masterfile)                         ! words in line is retrieved
    select case (trim(words%word_vec(1)))

	case ('coo_nodeinput')
	  select case (trim(words%word_vec(2)))
	  case ('hawc2')
	    write(0,*) 'Node inputs are read in hawc2 global coordinates'
        TBH_GLOBAL = reshape((/ 1.d0, 0.d0, 0.d0, 0.d0, 1.d0, 0.d0, 0.d0, 0.d0, 1.d0/),(/3,3/))
	  case ('bladed')
	    write(0,*) 'Node inputs are read in bladed global coordinates'
        TBH_GLOBAL = reshape((/ 0.d0, 1.d0, 0.d0, 1.d0, 0.d0, 0.d0, 0.d0, 0.d0,-1.d0/),(/3,3/))
	  case default
	    write(0,*) '*** ERROR *** coo_nodeinput coordinate not understood'
		err=.true.
	  end select

    case ('density_water')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  density_water=rea(1)

    case ('gravity')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  gravity=rea(1)

    case ('mwl')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  mwl=rea(1)

    case ('mudlevel')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  mudlevel=rea(1)

    case ('density_steel')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  rho=rea(1)

    case ('emod_steel')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  emod=rea(1)

    case ('gmod_steel')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  gmod=rea(1)

    case ('density_marinegrowth')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  mg_rho=rea(1)

    case ('thickness_marinegrowth')                                     
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
	  mg_thick=rea(1)


!   --------------------------------------------
    case ('')
      continue
!   --------------------------------------------
    case ('end')
      out=.true.
      ! Control of data given above is OK
      all_ok=.true.

      if (all_ok==.true.) then
        write(0,*) 'Properties commands read with succes'
      else
        write(0,*) '*** ERROR *** Not all needed properties commands present - error'
      endif
    case default
      write(0,*) '*** ERROR ***  in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    end select
    if (err==.true.) then
      write(0,*) '*** ERROR ***  in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    endif
  enddo
end subroutine properties_input
!
Subroutine body_input(masterfile)
use genout_tools

Type (Tbodydef),pointer			   :: b=>null()

Type (Tmasterfile), pointer :: masterfile

Type (Tline) :: line
Type (Tword) :: words

integer :: int
real*8, dimension(maxwords) :: rea

logical :: init                     , &
           out                      , &
           name_control            , &
           nmembers_control            , &
           memberid_control            , &
           all_ok, &
           err

init                          =.false.
out                           =.false.
name_control                  =.false.
nmembers_control                 =.false.
memberid_control               =.false.
err=.false.
all_ok=.false.

!  bodydef_counter=bodydef_counter+1
  call inc_bodies_array()
  b=>bodies(size(bodies))

  do while (out==.false.)
    call readline(masterfile,line,init)        ! line is read
    call getwords(line%line,words,masterfile)                         ! words in line is retrieved
    select case (trim(words%word_vec(1)))

    case ('name')                                     
      b%name=words%word_vec(2)
      name_control=.true.

    case ('nmembers')  
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
      b%nmembers=nint(rea(1))                                 
      nmembers_control=.true.

    case ('memberid')  
	  if (nmembers_control.eq..false.) then
	    write(0,*) '*** ERROR *** In preprocessor, nelem must be specified before elements command is given'
		err=.true.
		exit
	  endif
      call get_parameters(words,2,b%nmembers,rea,err)
      if (err==.true.) exit
	  allocate(b%memberid(b%nmembers))
      b%memberid(1:b%nmembers)=nint(rea(1:b%nmembers))
      memberid_control=.true.
!   --------------------------------------------
    case ('')
      continue
!   --------------------------------------------
    case ('end')
      out=.true.
      ! Control of data given above is OK
      all_ok=.true.
      if ((name_control          ==.false.).or. &
          (nmembers_control              ==.false.).or. &
          (memberid_control             ==.false.)) then
        all_ok=.false.
      endif

      if (all_ok==.true.) then
        write(0,*) 'Body commands read with succes'
      else
        write(0,*) '*** ERROR *** Not all needed body commands present - error'
      endif
    case default
      write(0,*) '*** ERROR ***  in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    end select
    if (err==.true.) then
      write(0,*) '*** ERROR ***  in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    endif
  enddo
end subroutine body_input
!
Subroutine fix_input(masterfile)
use genout_tools
use global_vars

Type (Tbodydef),pointer			   :: b=>null()
Type (Tfix0),pointer			   :: f0=>null()
Type (Tfix1),pointer			   :: f1=>null()

Type (Tmasterfile), pointer :: masterfile

Type (Tline) :: line
Type (Tword) :: words

integer :: int
real*8, dimension(maxwords) :: rea

logical :: init                     , &
           out                      , &
           name_control            , &
           nmembers_control            , &
           memberid_control            , &
           all_ok, &
           err

init                          =.false.
out                           =.false.
name_control                  =.false.
nmembers_control                 =.false.
memberid_control               =.false.
err=.false.
all_ok=.false.

  do while (out==.false.)
    call readline(masterfile,line,init)        ! line is read
    call getwords(line%line,words,masterfile)                         ! words in line is retrieved
    select case (trim(words%word_vec(1)))

    case ('fix0')                                     
	  call inc_fix0_array()
	  f0=>fix0(size(fix0))
      f0%b=>get_body_by_name(words%word_vec(2))

    case ('fix1')                                     
	  call inc_fix1_array()
	  f1=>fix1(size(fix1))
      f1%b1=>get_body_by_name(words%word_vec(2))
      f1%b2=>get_body_by_name(words%word_vec(3))
      call get_parameters(words,4,1,rea,err)
      if (err==.true.) exit
	  f1%epsdist=rea(1)

!   --------------------------------------------
    case ('')
      continue
!   --------------------------------------------
    case ('end')
      out=.true.
      ! Control of data given above is OK
      all_ok=.true.

      if (all_ok==.true.) then
        write(0,*) 'Fix commands read with succes'
      else
        write(0,*) '*** ERROR *** Not all needed fix commands present - error'
      endif
    case default
      write(0,*) '*** ERROR ***  in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    end select
    if (err==.true.) then
      write(0,*) '*** ERROR ***  in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    endif
  enddo
end subroutine fix_input
!
subroutine inc_members_array()
! subroutine that increases the size of nodes array by one and keeps the old values
Type (Tmember),dimension(:),pointer :: temp

integer           :: i

  allocate(temp(size(members)+1))
  if (size(members).gt.0) then
    do i=1,size(members)
      temp(i)=members(i)
    enddo
	deallocate(members)
	allocate(members(size(temp)))
	members=temp
  else
	allocate(members(1))
  endif

end subroutine
!
subroutine inc_nodes_array()
! subroutine that increases the size of nodes array by one and keeps the old values
Type (Tnode),dimension(:),pointer :: temp

integer           :: i

  allocate(temp(size(nodes)+1))
  if (size(nodes).gt.0) then
    do i=1,size(nodes)
      temp(i)=nodes(i)
    enddo
	deallocate(nodes)
	allocate(nodes(size(temp)))
	nodes=temp
  else
	allocate(nodes(1))
  endif

end subroutine
!
subroutine inc_bodies_array()
! subroutine that increases the size of bodies array by one and keeps the old values
Type (Tbodydef),dimension(:),pointer :: temp

integer           :: i

  allocate(temp(size(bodies)+1))
  if (size(bodies).gt.0) then
    do i=1,size(bodies)
      temp(i)=bodies(i)
    enddo
	deallocate(bodies)
	allocate(bodies(size(temp)))
	bodies=temp
  else
	allocate(bodies(1))
  endif

end subroutine
!
subroutine inc_fix0_array()
! subroutine that increases the size of wind_ramps%wind_ramp array by one and keeps the old values
Type (Tfix0),dimension(:),pointer :: temp

integer           :: i

  allocate(temp(size(fix0)+1))
  if (size(fix0).gt.0) then
    do i=1,size(fix0)
      temp(i)=fix0(i)
    enddo
	deallocate(fix0)
	allocate(fix0(size(temp)))
	fix0=temp
  else
	allocate(fix0(1))
  endif

end subroutine
!
subroutine inc_fix1_array()
! subroutine that increases the size of the fix1 array by one and keeps the old values
Type (Tfix1),dimension(:),pointer :: temp

integer           :: i

  allocate(temp(size(fix1)+1))
  if (size(fix1).gt.0) then
    do i=1,size(fix1)
      temp(i)=fix1(i)
    enddo
	deallocate(fix1)
	allocate(fix1(size(temp)))
	fix1=temp
  else
	allocate(fix1(1))
  endif

end subroutine
!
function get_body_by_name(name)
! input
character*(*) name
! output
Type (Tbodydef),pointer :: get_body_by_name
! locals
integer  :: i
logical  :: found
!
found=.false.
do i=1,size(bodies)
  if (trim(bodies(i)%name)==trim(name)) THEN
    get_body_by_name => bodies(i)
    found=.true.
    exit
  endif
enddo
!
if (found==.false.) then
  write(0,*) '*** ERROR *** Main body not found, name: ',trim(name)
  STOP
  nullify(get_body_by_name)
endif
!
END FUNCTION get_body_by_name
!
subroutine get_orientation_thetaz(tangent,theta_z,tsb)
use utils, only : vec_norm,crossprod,rotate
real*8, dimension(3), intent(in) :: tangent
real*8, intent(in)               :: theta_z      ! angle between the new x-vector (section) 
                                                 ! and the old x-vector (body) in the x-y (body) plane
                                                 ! It gives direct contribution to angle of attack
real*8,dimension(3,3),intent(inout) :: tsb       ! Transformation matrix from section to body coo.
real*8,save							:: eps=1.d-6

! Z axis is the tangent to the c2-def curve directed from root to tip
tsb(:,3)=vec_norm(tangent)

! orientation of local aero y-axis to blade coo. (without twist)
! special cases that would lead to division by zero or other errors
if (tsb(2,3).GE.(1.d0-eps)) then 
  tsb(1,3)=0.d0
  tsb(2,3)=1.d0
  tsb(3,3)=0.d0
  tsb(1,2)=0.d0
  tsb(2,2)=0.d0
  tsb(3,2)=-1.d0
else if (tsb(2,3).LE.(-1.d0+eps)) then 
  tsb(1,3)=0.d0
  tsb(2,3)=-1.d0
  tsb(3,3)=0.d0
  tsb(1,2)=0.d0
  tsb(2,2)=0.d0
  tsb(3,2)=1.d0
else if (tsb(1,3).GE.(1.d0-eps)) then
  tsb(1,3)=1.d0
  tsb(2,3)=0.d0
  tsb(3,3)=0.d0
  tsb(1,2)=0.d0
  tsb(2,2)=1.d0
  tsb(3,2)=0.d0
else if (tsb(1,3).LE.(-1.d0+eps)) then
  tsb(1,3)=-1.d0
  tsb(2,3)=0.d0
  tsb(3,3)=0.d0
  tsb(1,2)=0.d0
  tsb(2,2)=1.d0
  tsb(3,2)=0.d0
else   ! Normal situation (y axis is assumed to have no component i x direction (no twist)
  tsb(1,2)=0.0
  tsb(2,2)= tsb(3,3)/sqrt(tsb(2,3)**2+tsb(3,3)**2)
  tsb(3,2)=-tsb(2,3)/sqrt(tsb(2,3)**2+tsb(3,3)**2)
  if (tsb(2,2) .lt. 0.d0) then    ! Choose the solution with the highest y-coordinate of the y-axis
    tsb(2,2) = -tsb(2,2)
    tsb(3,2) = -tsb(3,2)
  endif
endif

! orientation of local x-axis to blade coo. (without twist)
tsb(:,1)=crossprod(tsb(:,2),tsb(:,3))
! Rotation of x and y around z with the twist angle
tsb(:,1)=rotate(tsb(:,1),tsb(:,3),theta_z)
tsb(:,2)=rotate(tsb(:,2),tsb(:,3),theta_z) 
end subroutine get_orientation_thetaz
!
!*************************
subroutine get_orientations()
!*************************
integer :: i
real*8,dimension(3) :: tangent
do i=1,size(members)
  tangent=nodes(members(i)%node2)%R-nodes(members(i)%node1)%R
  call get_orientation_thetaz(tangent,0.d0,members(i)%TLG)
  members(i)%TLG=MATMUL(TBH_GLOBAL,members(i)%TLG)
enddo
end subroutine get_orientations
!
!*************************
SUBROUTINE GENINPUT_2HAWC2()
!*************************
! locals
INTEGER :: i
DOUBLE PRECISION :: T(3,3)
! 
!! Transform triad
!DO i=1,NMEMBERS
!  T = RESHAPE((/MEMBERS(i)%ex, MEMBERS(i)%ey, MEMBERS(i)%ez/),(/3,3/))
!  T = MATMUL(TBH_GLOBAL,T)
!  T = MATMUL(T,TRANSPOSE(TBH_LOCAL))
!  MEMBERS(i)%ex = T(1:3,1)
!  MEMBERS(i)%ey = T(1:3,2)
!  MEMBERS(i)%ez = T(1:3,3)
!ENDDO
!
! Transform node positions
DO i=1,size(nodes)
  NODES(i)%r = MATMUL(TBH_GLOBAL, NODES(i)%r)
ENDDO
!
END SUBROUTINE
!
!*************************
SUBROUTINE GENINPUT_CHECK(body)
!*************************
! input
TYPE (TBODYDEF) :: body
! locals
INTEGER         :: i
TYPE (TMEMBER), POINTER :: m1, m2
INTEGER                 :: n1,n2
!
!! Check order
!IF (body%NMEMBERS.GT.1) THEN
!  m1 => MEMBERS(body%MEMBERID(1))
!  m2 => MEMBERS(body%MEMBERID(2))
!  IF (m1%NODE2.NE.m2%NODE1) body%BREORDER = .TRUE.   
!ENDIF
!!
!! Check nodes
!DO i=2,body%NMEMBERS
!  m1 => MEMBERS(body%MEMBERID(i-1))
!  m2 => MEMBERS(body%MEMBERID(i  ))
!  if (body%BREORDER) THEN
!    n1 = m1%NODE1
!	n2 = m2%NODE2
!  ELSE
!    n1 = m1%NODE2
!	n2 = m2%NODE1
!  ENDIF
!  IF (n1.NE.n2) THEN
!    WRITE(*,'("Error in node order in body: ",A)') body%NAME
!	STOP
!  ENDIF
!ENDDO
!
! Output which members have varying properties
WRITE(*,*)"*************************************************"
WRITE(*,*) "THE FOLLOWING MEMBERS HAVE VARYING PROPERTIES"
DO i=1,body%NMEMBERS
  m1 => MEMBERS(body%MEMBERID(i))
  if ((m1%t1.ne.m1%t2).OR.(m1%d1.ne.m1%d2)) WRITE(*,*) m1
ENDDO
WRITE(*,*)"*************************************************"

END SUBROUTINE
!
!*************************
SUBROUTINE GENINPUT_HTC(body, unit)
!*************************
! input
TYPE (TBODYDEF) :: body
INTEGER         :: unit
! locals
DOUBLE PRECISION :: R0(3)=0.d0,xc=0.d0,yc=0.d0,zc=0.d0
INTEGER          :: isec=0,nsec=0
CHARACTER*45            :: secformat1 = '("      sec  ",I2,F11.5,F11.5,F11.5," 0.0 ;")'
CHARACTER*70            :: secformat2 = '("      sec  ",I2,F11.5,F11.5,F11.5," 0.0 ;  NODE: ",I3," D=",F11.5)'
INTEGER                 :: i,j
DOUBLE PRECISION        :: r(3),dr(3), rl(3), Ttranspose(3,3)
DOUBLE PRECISION        :: d1, d2
TYPE (TNODE), POINTER   :: n1,n2
TYPE (TMEMBER), POINTER :: m
real*8					:: zero=0.d0
!
! Count nof section points (nodes)
nsec = 1
DO i=1,body%NMEMBERS
  nsec = nsec + MEMBERS(body%MEMBERID(i))%HTC_DIVISION
ENDDO
!
! Allocate local node positions
body%nsec = nsec
ALLOCATE(body%RLOCAL(3,nsec))
!
! Create transformation matrix (based on 1st member)
m => MEMBERS(body%MEMBERID(1))   ! points to first member
!Ttranspose(1,1:3) = m%EX
!Ttranspose(2,1:3) = m%EY
!Ttranspose(3,1:3) = m%EZ
Ttranspose=transpose(m%TLG)   ! tjul
!
! Store transformation matrix (Euler parameters) in body definition
CALL AMAT2EULER(m%TLG,body%euler)
!
! Get origin
IF (body%BREORDER) THEN
  n1 => NODES(m%NODE2)
ELSE
  n1 => NODES(m%NODE1)
ENDIF
R0 = n1%R
!
! Store origin
body%inipos = R0
!
! Write headers
WRITE(unit,*)"  begin main_body;"
WRITE(unit,'("    name        ",A," ;")') body%name(1:len_trim(body%name))   
WRITE(unit,*)"    type        timoschenko ;"
WRITE(unit,*)"    nbodies     1 ;"
WRITE(unit,*)"    gravity 9.80665 ; [m/s^2]"
WRITE(unit,*)"    node_distribution     c2_def ;"     
WRITE(unit,*)"    damping_posdef   0.0 0.0 0.0 4.24E-3 4.24E-3 1.25E-3 ;"
!WRITE(unit,*)"    damping_posdef   1.0E-6 1.0E-6 1.0E-6 1.0E-6 1.0E-6 1.0E-6 ;"
!WRITE(unit,*)"    damping   0.0E0 0.0E0 0.0E0 0.0E0 0.0E0 0.0E0 ;"
WRITE(unit,*)"    begin timoschenko_input;"
WRITE(unit,'("      filename ./data_ja/",A,".stc ;")') body%name(1:len_trim(body%name))
WRITE(unit,*)"      set 1 1 ;"
WRITE(unit,*)"    end timoschenko_input;"
WRITE(unit,*)"    begin c2_def;"
WRITE(unit,'("      nsec ",I2,";")') nsec
!
! Write 1st point
isec = 1
m => MEMBERS(body%MEMBERID(1))
IF (body%BREORDER) THEN
  n1 => NODES(m%NODE2)
  d1 = m%d2
ELSE
  n1 => NODES(m%NODE1)
  d1 = m%d1
ENDIF
WRITE(unit,secformat2) isec,xc,yc,zc,n1%ID,d1
body%RLOCAL(1:3,isec) = (/xc,yc,zc/)
!
! Write remainders
DO i=1,body%NMEMBERS
  m => MEMBERS(body%MEMBERID(i))
  IF (body%BREORDER) THEN
    n1 => NODES(m%NODE2)
    n2 => NODES(m%NODE1)
	d2 = m%d1
  ELSE
    n1 => NODES(m%NODE1)
    n2 => NODES(m%NODE2)
	d2 = m%d2
  ENDIF
  dr = (n2%r - n1%r)/m%HTC_DIVISION
  DO j=1,m%HTC_DIVISION
    r = n1%r + DBLE(j)*dr - R0
	rl = MATMUL(Ttranspose,r)
    isec = isec + 1
    if (j.eq.m%HTC_DIVISION) THEN 
!      WRITE(unit,secformat2) isec,rl(1),rl(2),rl(3),n2%ID,d2
      WRITE(unit,secformat2) isec,zero,zero,rl(3),n2%ID,d2
    else
!      WRITE(unit,secformat1) isec,rl(1),rl(2),rl(3)
      WRITE(unit,secformat1) isec,zero,zero,rl(3)
    endif	  
	body%RLOCAL(1:3,isec) = rl
  ENDDO
ENDDO
!
WRITE(unit,*)"    end c2_def ;"
WRITE(unit,*)"  end main_body;"
WRITE(unit,*)";"
!
END SUBROUTINE
!
!
!*************************
SUBROUTINE GENINPUT_ST(body, unit)
!*************************
! input
TYPE (TBODYDEF) :: body
INTEGER         :: unit
! locals
DOUBLE PRECISION :: R0(3)=0.d0,xc=0.d0,yc=0.d0,zc=0.d0
INTEGER          :: isec=0, nsec=0, set
CHARACTER*45            :: secformat = '("      sec  ",I2,F11.5,F11.5,F11.5," 0.0 ;")'
INTEGER                 :: i,j
DOUBLE PRECISION        :: r(3),dr(3), rl(3), Ttranspose(3,3)
TYPE (TNODE), POINTER   :: n1,n2
TYPE (TMEMBER), POINTER :: m
DOUBLE PRECISION, PARAMETER :: eps = 1.d-3
DOUBLE PRECISION            :: intpol
DOUBLE PRECISION            :: t,t1,t2
DOUBLE PRECISION            :: d,d1,d2
TYPE (TSTLINE)              :: sec
DOUBLE PRECISION            :: PI
DOUBLE PRECISION            :: set_factor
!
! Init
PI = 4.d0*DATAN(1.d0)
!
! Count nof st-lines
nsec = 0
DO i=1,body%NMEMBERS
  nsec = nsec + 1 + MEMBERS(body%MEMBERID(i))%HTC_DIVISION
ENDDO
!
! Get origin node of this body
m => MEMBERS(body%MEMBERID(1))
IF (body%BREORDER) THEN
  n1 => NODES(m%NODE2)
ELSE
  n1 => NODES(m%NODE1)
ENDIF
R0 = n1%R
!
! Write header
WRITE(unit,*)"2	Nof sets in this file, set #1 original structure, set#2 is stiff"
!
! Write set
set_factor = 1.d0
DO set = 1,2
  WRITE(unit,'("#",I1)') set
  WRITE(unit,*) "r	m	x_cg	y_cg	rix	riy	xshear	yshear	E	G	I_x	I_y	I_p	k_x	k_y	Ae	spitch_l	x_le_l	y_le_l"
  WRITE(unit,'("$1 ",I)') nsec																	

  DO i=1,body%NMEMBERS
    m => MEMBERS(body%MEMBERID(i))
    IF (body%BREORDER) THEN
      n1 => NODES(m%NODE2)
      n2 => NODES(m%NODE1)
      t1 = m%t2
      t2 = m%t1
	  d1 = m%d2
	  d2 = m%d1
    ELSE
      n1 => NODES(m%NODE1)
      n2 => NODES(m%NODE2)
	  t1 = m%t1
	  t2 = m%t2
	  d1 = m%d1
	  d2 = m%d2
    ENDIF
    dr = (n2%r - n1%r)
    DO j=0,m%ST_DIVISION
      intpol = (eps/2.d0 + (1.d0-eps)*DBLE(j)/m%ST_DIVISION)
      r = n1%r + intpol*dr - R0
	  zc = DSQRT(DOT_PRODUCT(r,r))
	  t = t1 + (t2-t1)*intpol
	  d = d1 + (d2-d1)*intpol
!
      sec.r = zc
	  sec%Ae = PI*((d/2.d0)**2 - (d/2.d0 - t)**2)
	  if (m%marine_growth) then
	    sec%m  = sec%Ae*rho+PI*((d/2.d0+mg_thick)**2 - (d/2.d0)**2)*mg_rho
	  else
	    sec%m  = sec%Ae*rho
	  endif
	  sec%I_x = PI/4.d0*((d/2.d0)**4 - (d/2.d0 - t)**4)
	  sec%I_y = sec%I_x
	  sec%I_p = sec%I_y + sec%I_x
	  sec%rix = DSQRT(sec%I_y/sec%Ae)
	  sec%riy = sec%rix
	  sec%E   = Emod*set_factor
	  sec%G   = Gmod*set_factor
      WRITE(unit,'(19E13.5)') sec
!
    ENDDO
  ENDDO
  set_factor = set_factor*1.d6
ENDDO
!
END SUBROUTINE
!
!******************************************
SUBROUTINE GENINPUT_ORIENTATION(body, unit)
!******************************************
! input
TYPE (TBODYDEF) :: body
INTEGER         :: unit
! locals

WRITE(unit,*) "    begin base ;"
WRITE(unit, '("      body           ",A," ;")') body%NAME(1:LEN_TRIM(body%NAME))
WRITE(unit, '("      inipos         ",3F10.5," ;")') body%inipos
WRITE(unit, '("      body_eulerpar  ",4F10.5," ;")') body%euler
WRITE(unit,*) "    ; TBG matrix ;"
WRITE(unit, '("     ;    ",3F10.5," ;")') members(body%memberid(1))%TLG(1,1:3)
WRITE(unit, '("     ;    ",3F10.5," ;")') members(body%memberid(1))%TLG(2,1:3)
WRITE(unit, '("     ;    ",3F10.5," ;")') members(body%memberid(1))%TLG(3,1:3)
WRITE(unit,*) "    end base ;"
WRITE(unit,*)";"
!
END SUBROUTINE
!
!******************************************
SUBROUTINE GENINPUT_FIX0(body, unit)
!******************************************
! input
TYPE (TBODYDEF) :: body
INTEGER         :: unit
! locals
WRITE(unit,*) "    begin fix0 ;"
WRITE(unit, '("      body   ",A," ;")') body%NAME(1:LEN_TRIM(body%NAME))
WRITE(unit,*) "    end fix0 ;"
WRITE(unit,*)";"
!
END SUBROUTINE
!
!*******************************************************
SUBROUTINE GENINPUT_FIX1(body1,body2,epsdist,unit)
!*******************************************************
! input
TYPE (TBODYDEF) :: body1, body2
DOUBLE PRECISION :: epsdist
INTEGER         :: unit
! locals
INTEGER         :: node1, node2
DOUBLE PRECISION :: A1(3,3),A2(3,3)
LOGICAL          :: bFound
DOUBLE PRECISION :: dist, mindist
DOUBLE PRECISION :: R1(3), R2(3)
!
CALL CAMAT(body1%EULER, A1)
CALL CAMAT(body2%EULER, A2)
bFound = .FALSE.
mindist = 1.d6
DO node1=1,body1%NSEC
  R1 = body1%INIPOS + MATMUL(A1,body1%RLOCAL(1:3,node1))
  DO node2=1,body2%NSEC
    R2 = body2%INIPOS + MATMUL(A2,body2%RLOCAL(1:3,node2))
	dist = DSQRT(DOT_PRODUCT(R1-R2,R1-R2))
	if (dist.lt.mindist) mindist = dist
	IF (dist.lt.epsdist) THEN
	  bFound = .TRUE.
	  EXIT
	ENDIF
  ENDDO
  IF (bFound) exit
ENDDO
!
IF (bFound) THEN
  WRITE(unit,*) "    begin fix1 ;"
  WRITE(unit, '("      mbdy1   ",A," ",I," ;")') body1%NAME(1:LEN_TRIM(body1%NAME)),node1
  WRITE(unit, '("      mbdy2   ",A," ",I," ;")') body2%NAME(1:LEN_TRIM(body2%NAME)),node2
  WRITE(unit,*) "    end fix1 ;"
  WRITE(unit,*) ";"
ELSE
  WRITE(*,'("Intersection between ",A," and ",A," was not found!")') body1%NAME(1:LEN_TRIM(body1%NAME)),body2%NAME(1:LEN_TRIM(body2%NAME))  
  WRITE(*,'("Minimum dist was : ",E)') mindist
ENDIF
!
END SUBROUTINE
!
!
!*************************
SUBROUTINE GENINPUT_HYDRO(body, unit)
!*************************
! input
TYPE (TBODYDEF) :: body
INTEGER         :: unit
! locals
INTEGER                 :: isec=0,nsec=0
CHARACTER*70            :: secformat = '("      sec  ",F8.3,"   1    1   ", F11.7,F13.7,F7.2,"  0.0 0.0 0.0 0.0 ",F13.7)'
INTEGER                 :: i
DOUBLE PRECISION        :: d1, d2
DOUBLE PRECISION        :: li
DOUBLE PRECISION        :: r0(3),ri(3)
DOUBLE PRECISION        :: PI
TYPE (TMEMBER), POINTER :: m
TYPE (TNODE), POINTER   :: n1,n2
!
character*2000,dimension(:),pointer :: strings     ! data is first written here, then later to file
integer								:: nstrings    ! number of used lines in strings
real*8								:: eps=1.d-2
TYPE (TMEMBER), POINTER				:: mnext,mprev
real*8								:: d_last
real*8								:: dif_last
!
! Init
PI = 4.d0*DATAN(1.d0)
!
! Count nof section points (nodes)
nsec = body%nsec
!
allocate(strings(nsec*2))
nstrings=0
!
! Write headers
WRITE(unit,*)"  begin HYDRO_ELEMENT ;"
WRITE(unit,'("    mbdy_name ",A," ;")') body%name(1:len_trim(body%name))   
WRITE(unit,*)"    buoyancy 1 ;"
WRITE(unit,*)"    update_states  1 ; (0: no dynamic interaction, 1: fully coupled solution"
WRITE(unit,*)"    hydrosections auto 4 ; dist, of hydro calculation points from 1 to nsec"
!WRITE(unit,'(A,I3,A)')"    nsec",nsec," ;  z      Cm   Cd         A            Aref   width    dr/dz Cd_a_(quad) Cm_a Cd_a_lin Aif"
!
! Write 1st point
isec   = 1
m => MEMBERS(body%MEMBERID(1))
IF (body%BREORDER) THEN
  n1 => NODES(m%NODE2)
  d1 = m%d2
ELSE
  n1 => NODES(m%NODE1)
  d1 = m%d1
ENDIF
if (m%marine_growth) then
  d1=d1+2.d0*mg_thick
else
  continue
endif
!WRITE(unit,10)"      sec  ",0.0,"   1    1   ", PI/4.D0*d1**2,PI/4.D0*d1**2,d1,"  0.0 0.0 0.0 0.0 ",PI/4.D0*m%Dif**2," ;"
nstrings=nstrings+1
d_last=d1
Dif_last=m%Dif
WRITE(strings(nstrings),10)"      sec  ",0.0,"   1    1   ", PI/4.D0*d1**2,PI/4.D0*d1**2,d1,"  0.0 0.0 0.0 0.0 ",PI/4.D0*m%Dif**2," ;"
r0 = n1%R
m=>MEMBERS(body%MEMBERID(body%NMEMBERS))
IF (body%BREORDER) THEN
  n2 => NODES(m%NODE1)
ELSE
  n2 => NODES(m%NODE2)
END IF
!
! Write remainders
DO i=1,body%NMEMBERS
  m => MEMBERS(body%MEMBERID(i))
  IF (body%BREORDER) THEN
    n2 => NODES(m%NODE1)
    d2 = m%d1
  ELSE
    n2 => NODES(m%NODE2)
    d2 = m%d2
  ENDIF
  ri = n2%R - r0
  li = dsqrt(ri(1)**2+ri(2)**2+ri(3)**2)
  if (m%marine_growth) then
    d2=d2+2.d0*mg_thick
  endif
  mprev=>members(body%MEMBERID(max(i-1,1)))
  mnext=>members(body%MEMBERID(min(i+1,body%nmembers)))
  if ((d2.ne.d_last).and.(m%dif.ne.Dif_last) &
      .or.(i.eq.body%nmembers) & 
      .or.(m%marine_growth.ne.mnext%marine_growth) &
	  .or.(m%marine_growth.ne.mprev%marine_growth) ) then
    ! only unique lines are written
    !WRITE(unit,10)"      sec  ",li,"   1    1   ", PI/4.D0*d2**2,PI/4.D0*d2**2,d2,"  0.0 0.0 0.0 0.0 ",PI/4.D0*m%Dif**2," ;"
	nstrings=nstrings+1
	WRITE(strings(nstrings),10)"      sec  ",li,"   1    1   ", PI/4.D0*d2**2,PI/4.D0*d2**2,d2,"  0.0 0.0 0.0 0.0 ",PI/4.D0*m%Dif**2," ;"
  endif
  d_last=d2
  Dif_last=m%Dif
  if (m%marine_growth.ne.mnext%marine_growth &
      .or.(m%dif.ne.mnext%dif)) then
    IF (body%BREORDER) THEN
      n1 => NODES(mnext%NODE2)
      d1 = m%d2
    ELSE
      n1 => NODES(mnext%NODE1)
      d1 = m%d1
    ENDIF
    ri = n1%R - r0
    li = dsqrt(ri(1)**2+ri(2)**2+ri(3)**2)+eps
    if (mnext%marine_growth) then
      d1=d1+2.d0*mg_thick
    endif
	nstrings=nstrings+1
	WRITE(strings(nstrings),10)"      sec  ",li,"   1    1   ", PI/4.D0*d1**2,PI/4.D0*d1**2,d1,"  0.0 0.0 0.0 0.0 ",PI/4.D0*mnext%Dif**2," ;"
  endif
ENDDO
!
WRITE(unit,'(A,I3,A)')"    nsec",nstrings," ;  z      Cm   Cd         A            Aref   width    dr/dz Cd_a_(quad) Cm_a Cd_a_lin Aif"
do i=1,nstrings
  write(unit,20) trim(strings(i))
enddo
strings=>null()
nstrings=0
WRITE(unit,*)"  end HYDRO_ELEMENT ;"
!
return
  10 format (A,F8.3,A,F11.7,F13.7,F7.2,A,F13.7,A)
  20 format (A)
END SUBROUTINE
!
END MODULE